

/* sonar_tempcorrected
 *  
 *  Hőmérsékletre kompenzált távolságmérés HC-SR04 ultrahangos szenzorral, 
 *  a NewPing programkönyvtár használatával. A távolságot 500 ms-onként
 *  mérjük, s a Multifunkciós kártya 7-szegmenses kijelzőjén
 *  méterben, 3 tizedesjegyre kiírjuk az eredményt.
 *  A távolság kiszámításánál figyelembe vesszük a hang terjedési 
 *  sebességének hőmérséklettől való függését.
 *  A hőmérsékletet az A4 analóg bemenetre csatlakoztatott MC9700
 *  analóg hőmérővel mérjük, a beépített 1.1 V-os referenciához 
 *  viszonyítva.
 */
#include <NewPing.h>

/* HC-SR04 paraméterek */
#define TRIGGER_PIN  5
#define ECHO_PIN     6
#define MAX_DISTANCE 200

/* Multifunkciós kártya kijelző vezérlés */
/* A kijelzőt vezérlő lábak definiálása */
#define LATCH_DIO 4
#define CLK_DIO   7
#define DATA_DIO  8
/* Szegmensrajzolat a 0-9 számjegyekhez */
const byte SEGMENT_MAP[] = {0xC0, 0xF9, 0xA4, 0xB0, 0x99, 0x92, 0x82, 0xF8, 0X80, 0X90};
/* Az 1-4 számjegyek kiválasztása */
const byte SEGMENT_SELECT[] = {0xF1, 0xF2, 0xF4, 0xF8};
volatile byte segment_data[] = {0, 0, 0, 0}; // adattároló buffer
byte idx = 0;                    // Kijelzés számlálója
//--- ADC méréshez használt változók ----------
volatile long adc_sum = 0;       // analog value
volatile long adc_val = 0;       // ADC 1100 mérés összege
uint16_t adc_cnt = 1100;         // ADC számláló
//--- Eseményvezérelt távolságméréshez használt változók ---
unsigned int pingSpeed = 500;    // Ilyen gyakorisággal mérünk (500 ms)
unsigned long pingTimer;         // Időbélyeg a következő méréshez
volatile long sonar_val = 0;     // A kiolvasott eredmény

NewPing sonar(TRIGGER_PIN, ECHO_PIN, MAX_DISTANCE);

void setup() {
  //--- Vezérlő lábak konfigurálása ---
  pinMode(LATCH_DIO, OUTPUT);
  pinMode(CLK_DIO, OUTPUT);
  pinMode(DATA_DIO, OUTPUT);
  pingTimer = millis() + 50;     // Ekkor kezdjük a mérést
  //--- ADC indítás megszakításos módban ---
  DIDR0 = 0x3F;                  // digital inputs disabled
  ADMUX = _BV(REFS0) | _BV(REFS1) | _BV(MUX2); // A4 csatorna, 1.1 V ref
  ADCSRA = _BV(ADEN) |           // ADC enable
           _BV(ADATE) |          // auto trigger enable
           _BV(ADIE) |           // interrupt enable
           _BV(ADPS2);           // prescaler = 16  (azaz 2^4)
  ADCSRB = 0x00;                 // AD channels MUX off, free running mode
  ADCSRA |= _BV(ADSC);           // Konverzió indítása
  sei();                         // megszakítás globális engedélyezése
}

void loop() {
  if (millis() >= pingTimer) {   // Ha elérkezett az idő
    pingTimer += pingSpeed;      // Következő mérési időpont előjegyzése
    sonar.ping_timer(echoCheck); // Mérés indítása, az echoCheck visszahívási függvény regisztrálása
  }
  if (sonar_val) {               // Ha van új adat
    long mv = adc_val / 1024;    // mv-ban mért feszültség
    float tempC = (mv-500)/10.0; // Hőmérsékletre átszámítjuk
    int distance = sonar_val * (331.5 + 0.6 * tempC) / 2000;      // milliméter egységben
    sonar_val = 0;               // Új adatot várunk bele
    segment_data[0] = SEGMENT_MAP[(distance / 1000) % 10] & 0x7F; // Tizedesponttal...
    segment_data[1] = SEGMENT_MAP[(distance / 100) % 10];
    segment_data[2] = SEGMENT_MAP[(distance  / 10) % 10];
    segment_data[3] = SEGMENT_MAP[distance  % 10];
  }
  refresh_display();
}

/* A soron következő számjegy megjelenítése */
void refresh_display() {
  byte i = idx++ & 0x3;
  digitalWrite(LATCH_DIO, LOW);
  shiftOut(DATA_DIO, CLK_DIO, MSBFIRST, segment_data[i]);
  shiftOut(DATA_DIO, CLK_DIO, MSBFIRST, SEGMENT_SELECT[i] );
  digitalWrite(LATCH_DIO, HIGH);
}

//--- Távolságmérés végére várunk -----------
void echoCheck() {                     // Timer2 megszakításból hívjuk
  if (sonar.check_timer()) {           // Így ellenőrizzük, hogy volt-e már válasz
    sonar_val = sonar.ping_result;     // Eredmény kiolvasása
  }
}

//--- 'ADC kész' megszakítás kiszolgálása ---
ISR(ADC_vect) {
  adc_sum += ADCW;         // 1100 mérést összegzünk
  if (--adc_cnt == 0) {
    adc_val = adc_sum;
    adc_sum = 0;
    adc_cnt = 1100;
  }
}
